    ".br.com",	"whois.centralnic.net",
    ".cn.com",	"whois.centralnic.net",
    ".eu.com",	"whois.centralnic.net",
    ".gb.com",	"whois.centralnic.net",
    ".gb.net",	"whois.centralnic.net",
    ".hu.com",	"whois.centralnic.net",
    ".no.com",	"whois.centralnic.net",
    ".qc.com",	"whois.centralnic.net",
    ".sa.com",	"whois.centralnic.net",
    ".se.com",	"whois.centralnic.net",
    ".se.net",	"whois.centralnic.net",
    ".uk.com",	"whois.centralnic.net",
    ".uk.net",	"whois.centralnic.net",
    ".us.com",	"whois.centralnic.net",
    ".uy.com",	"whois.centralnic.net",
    ".za.com",	"whois.centralnic.net",
    ".gov",	"whois.nic.gov",
    ".int",	"whois.isi.edu",
    ".mil",	"whois.nic.mil",
    ".ac",	"whois.nic.ac",
    ".ad",	"\003",
    ".ae",	"\001http://cc.emirates.net.ae/Customer_care/cc_card/check_domains.choose_domains/",
    ".af",	"whois.nic.af",
    ".ag",	"\001http://www.nic.ag/domain_search.htm",
    ".ai",	"\003",
    ".al",	"\003",
    ".am",	"whois.amnic.net",
    ".am",	"\001http://www.amnic.net/cgi/engine/db/whois.html",
    ".aq",	"\003",
    ".ar",	"\001http://www.nic.ar/consultas/consdom.htm",
    ".art",	"whois.nic.art",
    ".as",	"whois.nic.as",
    ".at",	"whois.ripe.net",
    ".net.au",	"whois.connect.com.au",
    ".au",	"whois.auda.org.au",
    ".ba",	"\003",
    ".bb",	"\001http://domains.org.bb/regsearch/",
    ".be",	"whois.ripe.net",
    ".bf",	"\003",
    ".bg",	"whois.ripe.net",
    ".bi",	"\001http://www.nic.cd/database/bi/",
    ".bm",	"\003",
    ".bo",	"\001http://www.nic.bo/listadom.phtml",
    ".br",	"whois.nic.br",
    ".bt",	"whois.nic.tm",
    ".bv",	"\003",
    ".bz",	"\003",
    ".ca",	"whois.cira.ca",
    ".cc",	"whois.nic.cc",
    ".cd",	"\001http://www.nic.cd/database/cd/",
    ".cg",	"\001http://www.nic.cd/database/cg/",
    ".ch",	"whois.nic.ch",
    ".ci",	"\003",
    ".ck",	"whois.nic.ck",
    ".cl",	"whois.nic.cl",
    ".cm",	"\003",
    ".ac.cn",	"whois.cnc.ac.cn",
    ".cn",	"whois.cnnic.net.cn",
    ".uk.co",	"whois.uk.co",
    ".co",	"\001http://daimon.uniandes.edu.co:8890/dominio/plsql/PConsulta.ConsultarDominio",
    ".cr",	"\001http://www.nic.cr/consulta-dns.html",
    ".cu",	"\001http://www.nic.cu/consultas/consult.html",
    ".cx",	"whois.nic.cx",
    ".cy",	"whois.ripe.net",
    ".cz",	"whois.nic.cz",
    ".de",	"whois.denic.de",
    ".dk",	"whois.dk-hostmaster.dk",
    ".do",	"\001http://www.nic.do",
    ".dz",	"whois.nic.dz",
    ".ec",	"whois.nic.ec",
    ".ee",	"\003",
    ".es",	"whois.nic.es",
    ".eu",	"whois.eu",
    ".fi",	"whois.fi",
    ".fj",	"whois.usp.ac.fj",
    ".fk",	"\003",
    ".fm",	"\001http://www.nic.fm/register.html",
    ".fo",	"whois.nic.fo",
    ".fr",	"whois.nic.fr",
    ".gb",	"\003",
    ".gd",	"whois.adamsnames.com",
    ".ge",	"\001http://www.nic.net.ge",
    ".gf",	"whois.nplus.gf",
    ".gg",	"whois.gg",
    ".gh",	"\003",
    ".gi",	"whois2.afilias-grs.net",
    ".gl",	"whois.nic.gl",
    ".gm",	"whois.ripe.net",
    ".gn",	"\003",
    ".gr",	"whois.ripe.net",
    ".gs",	"whois.nic.gs",
    ".gu",	"\001http://gadao.gov.gu/Scripts/wwsquery/wwsquery.dll?hois=guamquery",
    ".gy",	"whois.registry.gy",
    ".hk",	"whois.hkirc.hk",
    ".hm",	"\003",
    ".hn",	"whois2.afilias-grs.net",
    ".hr",	"whois.dns.hr",
    ".ht",	"whois.nic.ht",
    ".hu",	"whois.nic.hu",
    ".id",	"whois.idnic.net.id",
    ".ie",	"\001http://www.domainregistry.ie/search/whois.html",
    ".ie",	"whois.domainregistry.ie",
    ".il",	"whois.isoc.org.il",
    ".im",	"whois.nic.im",
    ".in",	"whois.registry.in",
    ".io",	"whois.nic.io",
    ".ir",	"whois.nic.ir",
    ".is",	"whois.isnic.is",
    ".it",	"whois.nic.it",
    ".je",	"whois.je",
    ".jo",	"\003",
    ".jp",	"whois.jprs.jp",
    ".ke",	"whois.kenic.or.ke",
    ".kg",	"whois.domain.kg",
    ".kr",	"whois.kr",
    ".kw",	"\001http://www.domainname.net.kw",
    ".ky",	"\003",
    ".kz",	"whois.nic.kz",
    ".la",	"whois.nic.la",
    ".lb",	"\001http://www.aub.edu.lb/lbdr/search.html",
    ".lc",	"\003",
    ".li",	"whois.nic.li",
    ".lk",	"whois.nic.lk",
    ".lr",	"\003",
    ".ls",	"\003",
    ".lt",	"whois.domreg.lt",
    ".lu",	"whois.dns.lu",
    ".lv",	"whois.nic.lv",
    ".ly",	"whois.nic.ly",
    ".ma",	"whois.iam.net.ma",
    ".mc",	"whois.ripe.net",
    ".md",	"whois.nic.md",
    ".me",	"whois.nic.me",
    ".mg",	"whois.nic.mg",
    ".mh",	"\003",
    ".mm",	"whois.nic.mm",
    ".mn",	"whois.nic.mn",
    ".mo",	"\003",
    ".mp",	"whois.nic.mp",
    ".mr",	"\003",
    ".ms",	"whois.nic.ms",
    ".mt",	"\001http://www.um.edu.mt/nic/dir/",
    ".mu",	"whois.nic.mu",
    ".mw",	"\001http://www.tarsus.net/whois/",
    ".mx",	"whois.mx",
    ".my",	"whois.domainregistry.my",
    ".na",	"whois.na-nic.com.na",
    ".nc",	"whois.nc",
    ".nf",	"whois.nic.net.nf",
    ".ng",	"whois.nic.net.ng",
    ".ni",	"\001http://ns.ni/whois.htm",
    ".nl",	"whois.domain-registry.nl",
    ".no",	"whois.norid.no",
    ".np",	"\001http://202.52.255.47/search.htm",
    ".nu",	"whois.nic.nu",
    ".nz",	"whois.srs.net.nz",
    ".org",	"whois.pir.org",
    ".pa",	"\001http://www.nic.pa",
    ".pe",	"kero.yachay.pe",
    ".pg",	"\003",
    ".ph",	"\001http://www.names.ph/search.html",
    ".pk",	"\003",
    ".pl",	"whois.dns.pl",
    ".pm",	"whois.nic.pm",
    ".pn",	"\003",
    ".pr",	"whois.nic.pr",
    ".ps",	"\003",
    ".pt",	"whois.dns.pt",
    ".pw",	"\003",
    ".py",	"\001http://www.nic.py/consultas/",
    ".qa",	"whois.registry.qa",
    ".re",	"whois.nic.re",
    ".ro",	"whois.rotld.ro",
    ".rs",	"whois.rnids.rs",
    ".ru",	"whois.tcinet.ru",
    ".rw",	"\001http://www.nic.cd/database/rw/",
    ".sa",	"whois.nic.net.sa",
    ".sb",	"whois.nic.net.sb",
    ".sc",	"whois2.afilias-grs.net",
    ".se",	"whois.iis.se",
    ".sg",	"whois.sgnic.sg",
    ".sh",	"whois.nic.sh",
    ".si",	"whois.arnes.si",
    ".sj",	"\003",
    ".sk",	"whois.sk-nic.sk",
    ".sm",	"whois.nic.sm",
    ".sn",	"\003",
    ".so",	"whois.nic.so",
    ".st",	"whois.nic.st",
    ".su",	"whois.tcinet.ru",
    ".sv",	"\001http://www.conacyt.gob.sv/dns/",
    ".sx",	"whois.sx",
    ".sz",	"\003",
    ".tc",	"whois.adamsnames.tc",
    ".tf",	"whois.nic.tf",
    ".tg",	"\001www.nic.tg",
    ".th",	"whois.thnic.co.th",
    ".tj",	"\003",
    ".tk",	"whois.dot.tk",
    ".tl",	"whois.nic.tl",
    ".tm",	"whois.nic.tm",
    ".tn",	"\003",
    ".to",	"whois.tonic.to",
    ".tp",	"\003",
    ".tr",	"whois.nic.tr",
    ".tt",	"\001http://www.nic.tt/cgi-bin/whois.cgi",
    ".tv",	"tvwhois.verisign-grs.com",
    ".tw",	"whois.twnic.net.tw",
    ".tz",	"whois.tznic.or.tz",
    ".ua",	"whois.ua",
    ".ug",	"\003",
    ".ac.uk",	"whois.ja.net",
    ".uk",	"whois.nic.uk",
    ".fed.us",	"whois.nic.gov",
    ".us",	"whois.nic.us",
    ".uy",	"whois.nic.org.uy",
    ".uz",	"whois.cctld.uz",
    ".va",	"whois.ripe.net",
    ".vc",	"whois2.afilias-grs.net",
    ".ve",	"whois.nic.ve",
    ".ve",	"\001http://www.nic.ve/nicwho01.html",
    ".vg",	"whois.adamsnames.tc",
    ".vn",	"\003",
    ".vu",	"\001http://www.vunic.vu/whois",
    ".wf",	"whois.nic.wf",
    ".ws",	"whois.samoanic.ws",
    ".yt",	"whois.nic.yt",
    ".za",	"\003",
    ".zm",	"\003",
    ".zr",	"\003",
    "-au-dom",	"whois.aunic.net",
    "-dom",	"whois.networksolutions.com",
    "-org",	"whois.networksolutions.com",
    "-hst",	"whois.networksolutions.com",
    "-arin",	"whois.arin.net",
    "-ripe",	"whois.ripe.net",
    "-mnt",	"whois.ripe.net",
    "-gandi",	"whois.gandi.net",
    "-ap",	"whois.apnic.net",
    "-au",	"whois.aunic.net",
    "-ti",	"whois.telstra.net",
    "-is",	"whois.isnet.is",
    "-6bone",	"whois.6bone.net",
    "-norid",	"whois.norid.no",
    "-ripn",	"whois.ripn.net",
    "-sgnic",	"whois.nic.net.sg",
    "-metu",	"whois.metu.edu.tr",
    "-cknic",	"whois.nic.ck",
    "-kg",	"whois.domain.kg",
